#!/user/bin/env python
# -*- coding: utf-8 -*-
"""
Copyright (C) 2024 Galaxy Technologies
Licensed under the Apache License, Version 2.0
"""
import numpy as np
from scipy.stats import norm
from pricelib.common.utilities.enums import UpDown, BarrierType, PaymentType, InOut, ExerciseType
from pricelib.common.processes import StochProcessBase
from pricelib.common.date_handler import global_evaluation_date
from pricelib.common.pricing_engine_base.analytic_engine_base import (AnalyticEngine, bs_formula,
                                                                      euro_asset_or_nothing, euro_cash_or_nothing)


class AnalyticBarrierEngine(AnalyticEngine):
    """障碍期权闭式解定价引擎
    欧式障碍期权：
        欧式香草与欧式资产或无、欧式现金或无的解析解的组合
    美式障碍期权：
        不支持同时输入非零的现金返还和不等于1的参与率，二者只有其一时支持。
        敲入期权现在支付期权费，但是当到期前资产价格触及障碍水平时，期权才生效。若一直没有发生敲入，则到期时支付现金返还rebate
        敲出期权现在支付期权费，但是当到期前资产价格触及障碍水平时，期权就失效了。若到期前发生敲出事件，则立刻支付现金返还rebate
        Merton(1973), Reiner & Rubinstein(1991a)提出障碍期权解析解，
        Broadie, Glasserman和Kou(1995)提出均匀离散观察障碍期权近似解"""

    def __init__(self, stoch_process: StochProcessBase = None, for_haug=False, *,
                 s=None, r=None, q=None, vol=None):
        """
        Args:
            stoch_process: 随机过程StochProcessBase对象
            for_haug: bool，Haug(1998)的双边障碍解析解，需要使用单边障碍期权解析解的公式直接计算出来的值，这种情况下不能应用
                            已敲入时payoff变为香草收益、或者已敲出时payoff变为现金返还的修正.
        在未设置stoch_process时，(stoch_process=None)，会默认创建BSMprocess，需要输入以下变量进行初始化
            s: float，标的价格
            r: float，无风险利率
            q: float，分红/融券率
            vol: float，波动率
        """
        super().__init__(stoch_process=stoch_process, s=s, r=r, q=q, vol=vol)
        self.for_haug = for_haug

    # pylint: disable=invalid-name, too-many-locals, missing-docstring, too-many-branches, too-many-statements
    def calc_present_value(self, prod, t=None, spot=None):
        """计算现值
        Args:
            prod: Product产品对象
            t: datetime.date，估值日; 如果是None，则使用全局估值日globalEvaluationDate
            spot: float，估值日标的价格，如果是None，则使用随机过程的当前价格
        Returns: float，现值
        """
        # 参数检查
        self._check_product_params(prod)

        calculate_date = global_evaluation_date() if t is None else t
        tau = prod.trade_calendar.business_days_between(calculate_date, prod.end_date) / prod.t_step_per_year  # 从估值日到到期日的交易日时间（年）
        # maturity = (prod.end_date - calculate_date).days / prod.annual_days.value  # 从估值日到到期日的自然日时间（年）
        tenor = (prod.stop_date - prod.start_date).days / prod.annual_days.value  # 从起始日到终止日的自然日时间（年）
        # 保底收益率
        _floor_yield = 0
        if prod.floor_yield != 0:  # 保底收益率需要输入期初价格s0作为名义本金
            _floor_yield = prod.floor_yield * tenor * prod.s0
        # 敲出/未敲入现金返还
        if prod.rebate_annual:
            _rebate = prod.rebate * tenor * prod.s0  # 年化百分比（仅适用于到期支付rebate的情形，敲出立即支付需自行按敲出时间计算）
        else:
            _rebate = prod.rebate  # 绝对数值
        if spot is None:
            spot = self.process.spot()
        r = self.process.interest(tau)
        q = self.process.div(tau)
        vol = self.process.vol(tau, self.process.spot())

        if calculate_date == prod.stop_date:  # 如果估值日就是终止日，则直接计算payoff
            value = self.calc_maturity_payoff(prod, np.array([spot]), _rebate, _floor_yield)[0]
            return value

        if prod.obs_type == ExerciseType.European:  # 如果是欧式障碍期权（到期观察）
            return self._calc_european_barrier(prod, spot, tau, r, q, vol, _rebate, _floor_yield)
        elif prod.obs_type == ExerciseType.American:  # 如果是美式障碍期权
            return self._calc_american_barrier(prod, spot, tau, r, q, vol, _rebate)
        else:
            raise ValueError("obs_type must be European or American")

    def _check_product_params(self, prod):
        """检查产品参数是否合法"""
        if prod.obs_type == ExerciseType.European:  # 如果是欧式障碍期权（到期观察）
            assert prod.payment_type == PaymentType.Expire, "ValueError: 欧式障碍期权（到期观察），payment_type只支持Expire到期支付"
        elif prod.obs_type == ExerciseType.American:  # 如果是美式障碍期权
            assert not prod.rebate_annual and prod.floor_yield == 0, "ValueError: 美式障碍解析解只支持绝对数值的现金返还，非百分比，非年化; 不支持保底收益率"
            assert prod.parti == 1 or prod.rebate == 0, "ValueError: 障碍期权解析解，由于公式限制，参与率parti不等于1、现金返还rebate不等于0不能共存，否则结果错误"
            if prod.inout == InOut.In:
                assert prod.payment_type == PaymentType.Expire, "ValueError: 敲入期权解析解，一直未敲入，到期时支付现金返还。payment_type只支持Expire"
            if prod.inout == InOut.Out:
                assert prod.payment_type == PaymentType.Hit, "ValueError: 敲出期权解析解，发生敲出，立刻支付现金返还。payment_type只支持Hit"

    @staticmethod
    def calc_maturity_payoff(prod, s_vec, _rebate, _floor_yield):
        """到期时的期权价值，在障碍价格内侧，默认设定未敲入，未敲出
        Args:
            prod: Product产品对象
            s_vec: np.ndarray, 网格的价格格点
            _rebate: float, 现金补偿绝对数值 (若为年化，计入时间为从起始日到终止日)
            _floor_yield: float, 保底收益绝对数值，计入时间为从起始日到终止日
        Returns:
            payoff: np.ndarray, 期末价值向量
        """
        if prod.inout == InOut.Out:
            v_vec = np.maximum(prod.callput.value * (s_vec - prod.strike), 0) * prod.parti + _floor_yield
            if prod.updown == UpDown.Up:
                v_vec = np.where(s_vec > prod.barrier, _rebate, v_vec)
            elif prod.updown == UpDown.Down:
                v_vec = np.where(s_vec < prod.barrier, _rebate, v_vec)
            else:
                raise ValueError("updown must be Up or Down")
        elif prod.inout == InOut.In:
            v_vec = _rebate * np.ones(len(s_vec))
            if prod.updown == UpDown.Up:
                v_vec = np.where(s_vec > prod.barrier, np.maximum(prod.callput.value * (s_vec - prod.strike),
                                                                  0) * prod.parti + _floor_yield, v_vec)
            elif prod.updown == UpDown.Down:
                v_vec = np.where(s_vec < prod.barrier, np.maximum(prod.callput.value * (s_vec - prod.strike),
                                                                  0) * prod.parti + _floor_yield, v_vec)
            else:
                raise ValueError("updown must be Up or Down")
        else:
            raise ValueError("inout must be In or Out")
        return v_vec

    def _calc_european_barrier(self, prod, spot, tau, r, q, vol, _rebate=0., _floor_yield=0.):
        """欧式障碍期权的解析解
            欧式香草与欧式资产或无、欧式现金或无的解析解的组合
            欧式现金或无的解析解可用于复制现金返还。
        Args:
            prod: Product产品对象
            spot: float，估值日标的价格
            tau: float，交易日到期时间（年）
            r: float，无风险利率（常数）
            q: float，连续分红率（常数）
            vol: float，波动率（常数）
            _rebate: float, 现金补偿绝对数值 (若为年化，计入时间为从起始日到终止日)
            _floor_yield: float, 保底收益绝对数值，计入时间为从起始日到终止日
        Returns:
            欧式障碍期权的解析解
        """
        if prod.barrier_type in [BarrierType.UIC, BarrierType.DOC]:  # 向上敲入看涨 或 向下敲出看涨
            if prod.barrier <= prod.strike:  # 此时相当于call的价值
                value = bs_formula(s=spot, k=prod.strike, t=tau, r=r, q=q, sigma=vol, sign=1) * prod.parti
            else:  # 此时 v = 方向向上、执行价为障碍价的 [ 资产或无 - 执行价 * 现金或无 ]
                value = (euro_asset_or_nothing(s=spot, k=prod.barrier, t=tau, r=r, q=q, sigma=vol, sign=1) -
                         euro_cash_or_nothing(s=spot, k=prod.barrier, t=tau, r=r, q=q, sigma=vol, sign=1) * prod.strike
                         ) * prod.parti
            # 现金返还rebate相当于看跌的现金或无; 如果存在保底收益率，则保底收益相当于看涨的现金或无.
            value += self._extra_returns(prod, spot, tau, r, q, vol, _rebate, _floor_yield, flag=1)
        elif prod.barrier_type in [BarrierType.DIC, BarrierType.UOC]:  # 向下敲入看涨 或 向上敲出看涨
            if prod.barrier <= prod.strike:  # 此时价值是0
                value = 0
            else:  # 此时 v = 多一份 call；空1份 方向向上、执行价为障碍价的 [ 资产或无 - 执行价 * 现金或无 ]
                value = (bs_formula(s=spot, k=prod.strike, t=tau, r=r, q=q, sigma=vol, sign=1) - (
                        euro_asset_or_nothing(s=spot, k=prod.barrier, t=tau, r=r, q=q, sigma=vol, sign=1) -
                        euro_cash_or_nothing(s=spot, k=prod.barrier, t=tau, r=r, q=q, sigma=vol, sign=1) * prod.strike
                )) * prod.parti
            # 现金返还rebate相当于看涨的现金或无; 如果存在保底收益率，则保底收益相当于看跌的现金或无.
            value += self._extra_returns(prod, spot, tau, r, q, vol, _rebate, _floor_yield, flag=-1)
        elif prod.barrier_type in [BarrierType.UIP, BarrierType.DOP]:  # 向上敲入看跌 或 向下敲出看跌
            if prod.barrier >= prod.strike:  # 此时价值是0
                value = 0
            else:  # 此时 v = 多一份 put；空1份 方向向下、执行价为障碍价的 [ 执行价 * 现金或无 - 资产或无 ]
                value = (bs_formula(s=spot, k=prod.strike, t=tau, r=r, q=q, sigma=vol, sign=-1) - (
                        euro_cash_or_nothing(s=spot, k=prod.barrier, t=tau, r=r, q=q, sigma=vol, sign=-1) * prod.strike
                        - euro_asset_or_nothing(s=spot, k=prod.barrier, t=tau, r=r, q=q, sigma=vol, sign=-1)
                )) * prod.parti
            # 现金返还rebate相当于看跌的现金或无; 如果存在保底收益率，则保底收益相当于看涨的现金或无.
            value += self._extra_returns(prod, spot, tau, r, q, vol, _rebate, _floor_yield, flag=1)
        elif prod.barrier_type in [BarrierType.DIP, BarrierType.UOP]:  # 向下敲入看跌 或 向上敲出看跌
            if prod.barrier >= prod.strike:  # 此时相当于put的价值
                value = bs_formula(s=spot, k=prod.strike, t=tau, r=r, q=q, sigma=vol, sign=-1) * prod.parti
            else:  # 此时 v = 方向向下、执行价为障碍价的 [ 执行价 * 现金或无 - 资产或无 ]
                value = (euro_cash_or_nothing(s=spot, k=prod.barrier, t=tau, r=r, q=q, sigma=vol, sign=-1) * prod.strike
                         - euro_asset_or_nothing(s=spot, k=prod.barrier, t=tau, r=r, q=q, sigma=vol, sign=-1)
                         ) * prod.parti
            # 现金返还rebate相当于看涨的现金或无; 如果存在保底收益率，则保底收益相当于看跌的现金或无.
            value += self._extra_returns(prod, spot, tau, r, q, vol, _rebate, _floor_yield, flag=-1)
        else:
            raise ValueError("Invalid barrier type")
        return value

    def _extra_returns(self, prod, spot, tau, r, q, vol, _rebate, _floor_yield, flag=1):
        """处理欧式障碍期权可能存在的额外payoff：现金返还rebate、保底收益率floor_yield
        Args:
            prod: Product产品对象
            spot: float，估值日标的价格
            tau: float，交易日到期时间（年）
            r: float，无风险利率（常数）
            q: float，连续分红率（常数）
            vol: float，波动率（常数）
            flag: int, 若 flag = 1， 现金返还rebate相当于看跌的现金或无; 如果存在保底收益率，则保底收益相当于看涨的现金或无.
                       若 flag = -1， 现金返还rebate相当于看涨的现金或无; 如果存在保底收益率，则保底收益相当于看跌的现金或无.
        Returns:

        """
        extra = 0
        if prod.floor_yield != 0:  # 保底收益率
            extra += euro_cash_or_nothing(s=spot, k=prod.barrier, t=tau, r=r, q=q, sigma=vol, sign=flag) * _floor_yield
        if prod.rebate != 0:  # 现金返还
            extra += euro_cash_or_nothing(s=spot, k=prod.barrier, t=tau, r=r, q=q, sigma=vol, sign=-flag) * _rebate
        return extra

    def _calc_american_barrier(self, prod, spot, tau, r, q, vol, _rebate):
        """美式障碍期权的解析解
            Merton(1973), Reiner & Rubinstein(1991a)提出障碍期权解析解，公式中的现金返还是绝对数值，不支持保底收益率floor_yield
            支持Broadie, Glasserman和Kou(1995)提出的均匀离散观察障碍期权近似解.
        Args:
            prod: Product产品对象
            spot: float，估值日标的价格
            tau: float，交易日到期时间（年）
            r: float，无风险利率（常数）
            q: float，连续分红率（常数）
            vol: float，波动率（常数）
            _rebate: float, 现金补偿绝对数值 (若为年化，计入时间为从起始日到终止日)
        Returns:
            美式障碍期权的解析解
        """
        drift = r - q
        if not self.for_haug:
            # 如果估值时的标的价格已经触碰障碍价，直接返回现金返还或者香草期权的价值
            if (prod.updown == UpDown.Up and spot >= prod.barrier) or (
                    prod.updown == UpDown.Down and spot <= prod.barrier):
                if prod.inout == InOut.Out:  # 发生敲出，立刻支付现金返还rebate
                    return prod.rebate
                elif prod.inout == InOut.In:  # 发生敲入，返回香草期权的价值
                    return bs_formula(s=spot, k=prod.strike, t=tau, r=r, q=q, sigma=vol, sign=prod.callput.value)
                else:
                    raise ValueError("inout只能是InOut.In或InOut.Out")

        if prod.discrete_obs_interval is not None:
            # 均匀离散观察障碍期权，M. Broadie, P. Glasserman, S.G. Kou(1997) 在连续障碍期权解析解上加调整项，调整障碍价格水平
            # 指数上的beta = -zeta(1/2) / sqrt(2*pi) = 0.5826, 其中zeta是黎曼zeta函数
            if prod.updown == UpDown.Up:  # 向上
                barrier = prod.barrier * np.exp(0.5826 * vol * np.sqrt(prod.discrete_obs_interval))
            elif prod.updown == UpDown.Down:  # 向下
                barrier = prod.barrier * np.exp(-0.5826 * vol * np.sqrt(prod.discrete_obs_interval))
            else:
                raise ValueError("self.updown 只能是Up或Down")
        else:  # 连续观察障碍期权
            barrier = prod.barrier

        mu = drift / vol ** 2 - 0.5
        la = np.sqrt(mu ** 2 + 2 * r / vol ** 2)
        a = (barrier / spot) ** (2 * mu)
        b = (barrier / spot) ** (2 * mu + 2)
        c = (barrier / spot) ** (mu + la)
        d = (barrier / spot) ** (mu - la)
        a1 = np.log(spot / prod.strike)
        a2 = np.log(spot / barrier)
        a3 = np.log(spot * prod.strike / barrier ** 2)
        a4 = drift + 0.5 * vol ** 2
        a5 = drift - 0.5 * vol ** 2
        a6 = tau
        a7 = vol * np.sqrt(a6)
        d1 = (a1 + a4 * a6) / a7
        d2 = (a1 + a5 * a6) / a7
        d3 = (a2 + a4 * a6) / a7
        d4 = (a2 + a5 * a6) / a7
        d5 = (a2 - a5 * a6) / a7
        d6 = (a2 - a4 * a6) / a7
        d7 = (a3 - a5 * a6) / a7
        d8 = (a3 - a4 * a6) / a7
        d9 = -a2 / a7 + la * a7
        d10 = -a2 / a7 - la * a7

        def A(phi):
            return phi * spot * np.exp(-q * a6) * norm.cdf(phi * d1) - \
                phi * prod.strike * np.exp(-r * a6) * norm.cdf(phi * d2)

        def B(phi):
            return phi * spot * np.exp(-q * a6) * norm.cdf(phi * d3) - \
                phi * prod.strike * np.exp(-r * a6) * norm.cdf(phi * d4)

        def C(phi, eta):
            return phi * spot * np.exp(-q * a6) * b * norm.cdf(-eta * d8) - \
                phi * prod.strike * np.exp(-r * a6) * a * norm.cdf(-eta * d7)

        def D(phi, eta):
            return phi * spot * np.exp(-q * a6) * b * norm.cdf(-eta * d6) - \
                phi * prod.strike * np.exp(-r * a6) * a * norm.cdf(-eta * d5)

        def E(eta):
            return prod.rebate * np.exp(-r * a6) * (norm.cdf(eta * d4) - a * norm.cdf(-eta * d5))

        def F(eta):
            return prod.rebate * (c * norm.cdf(eta * d9) + d * norm.cdf(eta * d10))

        if prod.barrier_type == BarrierType.UIC:  # 向上敲入看涨
            phi = 1
            eta = -1
            if prod.strike >= barrier:
                price = A(phi) + E(eta)
            else:
                price = (B(phi) - C(phi, eta) + D(phi, eta)) + E(eta)
        elif prod.barrier_type == BarrierType.UIP:  # 向上敲入看跌
            phi = -1
            eta = -1
            if prod.strike >= barrier:
                price = (A(phi) - B(phi) + D(phi, eta)) + E(eta)
            else:
                price = C(phi, eta) + E(eta)
        elif prod.barrier_type == BarrierType.UOC:  # 向上敲出看涨
            phi = 1
            eta = -1
            if prod.strike >= barrier:
                price = F(eta)
            else:
                price = (A(phi) - B(phi) + C(phi, eta) - D(phi, eta)) + F(eta)
        elif prod.barrier_type == BarrierType.UOP:  # 向上敲出看跌
            phi = -1
            eta = -1
            if prod.strike >= barrier:
                price = (B(phi) - D(phi, eta)) + F(eta)
            else:
                price = (A(phi) - C(phi, eta)) + F(eta)
        elif prod.barrier_type == BarrierType.DIC:  # 向下敲入看涨
            phi = 1
            eta = 1
            if prod.strike >= barrier:
                price = C(phi, eta) + E(eta)
            else:
                price = (A(phi) - B(phi) + D(phi, eta)) + E(eta)
        elif prod.barrier_type == BarrierType.DIP:  # 向下敲入看跌
            phi = -1
            eta = 1
            if prod.strike >= barrier:
                price = (B(phi) - C(phi, eta) + D(phi, eta)) + E(eta)
            else:
                price = A(phi) + E(eta)
        elif prod.barrier_type == BarrierType.DOC:  # 向下敲出看涨
            phi = 1
            eta = 1
            if prod.strike >= barrier:
                price = (A(phi) - C(phi, eta)) + F(eta)
            else:
                price = (B(phi) - D(phi, eta)) + F(eta)
        elif prod.barrier_type == BarrierType.DOP:  # 向下敲出看跌
            phi = -1
            eta = 1
            if prod.strike >= barrier:
                price = (A(phi) - B(phi) + C(phi, eta) - D(phi, eta)) + F(eta)
            else:
                price = F(eta)
        else:
            raise ValueError("self.barrier_type 障碍类型错误")
        return price * prod.parti



#!/user/bin/env python
# -*- coding: utf-8 -*-
"""
Copyright (C) 2024 Galaxy Technologies
Licensed under the Apache License, Version 2.0
"""
import datetime
from contextlib import suppress
import numpy as np
from pricelib.common.utilities.enums import (CallPut, InOut, UpDown, BarrierType, PaymentType, EngineType, StatusType,
                                             ExerciseType)
from pricelib.common.date_handler import CN_CALENDAR, AnnualDays, global_evaluation_date
from pricelib.common.utilities.patterns import Observer, global_auto_update_status
from pricelib.common.utilities.utility import time_this, logging
from pricelib.common.product_base.option_base import OptionBase
from pricelib.pricing_engines.analytic_engines import AnalyticVanillaEuEngine, AnalyticBarrierEngine
from pricelib.pricing_engines.fdm_engines import FdmVanillaEngine, FdmBarrierEngine
from pricelib.pricing_engines.mc_engines import MCBarrierEngine
from pricelib.pricing_engines.integral_engines import QuadBarrierEngine
from pricelib.pricing_engines.tree_engines import BiTreeBarrierEngine
from pricelib.products.vanilla.vanilla_option import VanillaOption



class BarrierOption(OptionBase, Observer):
    """单边障碍期权
    敲入期权在障碍被触及时才会生效，敲出期权在障碍被触及时立即失效"""

    def __init__(self, strike, barrier, rebate, callput=CallPut.Call, inout=InOut.Out, updown=UpDown.Up, *, parti=1,
                 obs_type=ExerciseType.American, payment_type=None, rebate_annual=False, floor_yield=0.0, engine=None,
                 status=StatusType.NoTouch, maturity=None, start_date=None, end_date=None, discrete_obs_interval=None,
                 trade_calendar=CN_CALENDAR, annual_days=AnnualDays.N365, t_step_per_year=243, stoch_process=None,
                 s0=None, s=None, r=None, q=None, vol=None, engine_method=None):
        """构造函数
        产品参数:
            strike: float, 执行价
            barrier: float, 障碍价
            rebate: float, 敲出现金返还金额，或未敲入现金返还金额
            rebate_annual: bool, 现金返还是否为年化，默认为False，即绝对数值，非年化，非百分比；若为True，则为百分比年化收益率
            callput: 看涨看跌类型，CallPut枚举类，Call看涨/Put看跌
            inout: 敲入敲出类型，InOut枚举类，In敲入/Out敲出
            updown: 向上向下类型，UpDown枚举类，Up向上/Down向下
            obs_type: 观察类型，ExerciseType枚举类，默认为American美式观察, 可选European欧式观察 (即到期观察)
            payment_type: 现金返还支付类型，PaymentType枚举类，敲入默认为到期支付Expire；敲出默认为立即支付Hit
            parti: float, 香草期权的参与率，默认为1
            floor_yield: float, 敲入或未敲出时，在香草的payoff之外，额外的保底收益率，年化收益，百分比小数
            status: 估值日前的敲入敲出状态，StatusType枚举类，默认为NoTouch未敲入/未敲出，UpTouch为向上敲入/敲出，DownTouch为向下敲入/敲出
        时间参数: 要么输入年化期限，要么输入起始日和到期日
            maturity: float，年化期限
            start_date: datetime.date，起始日
            end_date: datetime.date，到期日
            trade_calendar: 交易日历，Calendar类，默认为中国内地交易日历
            annual_days: int，每年的自然日数量
            t_step_per_year: int，每年的交易日数量
            discrete_obs_interval: 观察时间间隔. 若为连续观察，None；若为均匀离散观察，为年化的观察时间间隔
        可选参数:
            s0: float，标的期初价格，只有当 rebate_annual = True，即现金返还为年化收益率时；
                                  或者当 floor_yield != 0，即存在非零的保底收益率时，
                       需要输入标的期初价格s0，用于将百分比的收益率转化成绝对数值
            engine: 定价引擎，PricingEngine类，以下几种引擎均支持向上/向下、敲入/敲出、看涨/看跌8种单边障碍期权，支持敲出/未敲入现金返还
                            对于美式障碍，解析解和PDE引擎支持连续观察，其余引擎只支持离散观察(默认为每日观察)
                    解析解: AnalyticBarrierEngine 支持欧式障碍（到期观察）和美式障碍（支持连续观察/离散观察）
                                        欧式障碍：到期观察，支持保底收益率，支持支持现金返还为绝对数值/年化百分比
                                        美式障碍：敲入现金返还为到期支付；敲出现金返还为立即支付。现金返还为固定绝对数值，不支持年化的现金返还。
                                                不支持同时输入非零的现金返还和不等于1的参与率，二者只有其一时支持。不支持保底收益率
                    蒙特卡洛: MCBarrierEngine 支持欧式障碍（到期观察）和美式障碍（仅支持离散观察）
                                        欧式障碍：到期观察，支持保底收益率，支持支持现金返还为绝对数值/年化百分比
                                        美式障碍 ：敲入现金返还为到期支付；敲出现金返还 支持 立即支付/到期支付
                                                支持现金返还为绝对数值/年化百分比，支持保底收益率
                    PDE: FdmBarrierEngine  支持欧式障碍（到期观察）和美式障碍（支持连续观察/离散观察）
                                        欧式障碍：到期观察，支持保底收益率，支持支持现金返还为绝对数值/年化百分比
                                        美式障碍 ：敲入现金返还为到期支付；敲出现金返还 支持 立即支付/到期支付
                                                支持现金返回为绝对数值/年化百分比，支持保底收益率
                    积分法: QuadBarrierEngine  仅支持离散观察的美式障碍，todo: 支持欧式障碍、年化的现金返还、保底收益率
                                            敲入现金返还为到期支付；敲出现金返还 支持 立即支付/到期支付
                    二叉树: BiTreeBarrierEngine  仅支持离散观察的美式障碍，todo: 支持欧式障碍、年化的现金返还、保底收益率
                                            敲入现金返还为到期支付；敲出现金返还 支持 立即支付/到期支付
            engine_method: str, 定价引擎种类。若未输入engine，可以用engine_method快捷生成指定定价引擎。可选Analytic、MC/MonteCarlo、PDE/FDM、QUAD、TREE。
            若未提供引擎engine、也未指定引擎种类engine_method的情况下，如果提供了 随机过程 或 标的价格、无风险利率、分红/融券率、波动率，
            则默认使用解析解定价引擎 AnalyticBarrierEngine - Merton(1973), Reiner & Rubinstein(1991a)
                                支持 Broadie, Glasserman和Kou(1995)均匀离散观察调整
            stoch_process: 随机过程StochProcessBase对象
            s: float，标的价格
            r: float，无风险利率
            q: float，分红/融券率
            vol: float，波动率
        """
        super().__init__()
        self.trade_calendar = trade_calendar
        self.annual_days = annual_days
        self.t_step_per_year = t_step_per_year
        assert maturity is not None or (start_date is not None and end_date is not None), "Error: 到期时间或起止时间必须输入一个"
        self.start_date = start_date if start_date is not None else global_evaluation_date()  # 起始时间
        self.end_date = end_date if end_date is not None else (
                self.start_date + datetime.timedelta(days=round(maturity * annual_days.value)))  # 结束时间
        if self.trade_calendar.isholiday(self.end_date):  # 如果到期日不是交易日，向后顺延一个交易日
            self.end_date = self.trade_calendar.advance(self.end_date, datetime.timedelta(days=1))
        self.maturity = maturity if maturity is not None else (self.end_date - self.start_date).days / annual_days.value
        self.strike = strike
        self.barrier = barrier
        self.rebate = rebate  # 敲出现金返还，或未敲入现金返还
        self.rebate_annual = rebate_annual  # 现金返还是否为年化
        self.parti = parti
        self._callput = callput
        self._inout = inout
        self._updown = updown
        self.barrier_type = BarrierType.get_category(self._updown, self._inout, self._callput)
        if payment_type is None:  # 敲入默认为到期支付Expire；敲出默认为立即支付Hit
            if self._inout == InOut.In:
                self.payment_type = PaymentType.Expire
            else:
                self.payment_type = PaymentType.Hit
        else:
            self.payment_type = payment_type
            if self._inout == InOut.In:
                assert payment_type == PaymentType.Expire, "ValueError: 敲入期权的现金返还方式PaymentType一定是到期支付Expire"
        self.obs_type = obs_type
        self.floor_yield = floor_yield
        self.s0 = s0
        if self.rebate_annual or self.floor_yield != 0:
            assert s0 is not None, "Error: 现金返还为年化收益率时；或者存在非零的保底收益率时，需要输入标的期初价格s0，用于将百分比的收益率转化成绝对数值"
        self.discrete_obs_interval = discrete_obs_interval  # 连续观察=None；均匀离散观察=观察时间间隔
        if self.obs_type == ExerciseType.European:
            err_msg = ""
            if self.payment_type != PaymentType.Expire:
                err_msg += "ValueError: 到期观察的障碍期权，现金返还的支付方式只能是到期支付Expire.\n"
            if self.discrete_obs_interval is not None:
                err_msg += "ValueError: 到期观察的障碍期权，观察时间间隔只能为None.\n"
            if err_msg != "":
                raise ValueError(err_msg)
        self.status = status
        self.stop_date = self.end_date  # 停止日期，默认等于到期日。如果发生敲出并立即支付现金返还，停止日期更新为敲出日期
        if engine is not None:
            self.set_pricing_engine(engine)
        else:
            if engine_method is None:
                if self.analytic_check():
                    default_engine = AnalyticBarrierEngine
                else:
                    default_engine = FdmBarrierEngine
            elif engine_method == "Analytic":
                default_engine = AnalyticBarrierEngine
            elif engine_method in ["PDE", "FDM"]:
                default_engine = FdmBarrierEngine
            elif engine_method in ["MC", "MonteCarlo"]:
                default_engine = MCBarrierEngine
            elif engine_method == "QUAD":
                default_engine = QuadBarrierEngine
            elif engine_method == "TREE":
                default_engine = BiTreeBarrierEngine
            else:
                raise NotImplementedError(
                    f"不支持的engine_method{engine_method}, {self.__repr__()}仅支持Analytic/MC/PDE/QUAD/TREE")
            if stoch_process is not None:
                self.set_pricing_engine(default_engine(stoch_process))
            elif s is not None and r is not None and q is not None and vol is not None:
                self.set_pricing_engine(default_engine(s=s, r=r, q=q, vol=vol))

    def analytic_check(self):
        """检查输入的产品参数是否可以使用解析解定价引擎"""
        if self.obs_type == ExerciseType.American:  # 美式障碍期权解析解有限制条件
            if self.rebate_annual or self.floor_yield != 0:
                return False  # 美式障碍解析解只支持绝对数值的现金返还，非百分比，非年化; 不支持保底收益率
            if self.parti != 1 and self.rebate != 0:
                # 障碍期权解析解，由于公式限制，参与率parti不等于1、现金返还rebate不等于0不能共存，否则结果错误
                return False
            if self.inout == InOut.In and self.payment_type != PaymentType.Expire:
                # 敲入期权解析解，一直未敲入，到期时支付现金返还。payment_type只支持Expire
                return False
            if self.inout == InOut.Out and self.payment_type != PaymentType.Hit:
                # 敲出期权解析解，发生敲出，立刻支付现金返还。payment_type应该是Hit
                return False
        return True

    def set_pricing_engine(self, engine):
        """设置定价引擎，同时将自己注册为观察者。若已有定价引擎，先将自己从原定价引擎的观察者列表中移除"""
        with suppress(AttributeError, ValueError):
            self.engine.process.spot.remove_observer(self)
        self.engine = engine
        self.engine.process.spot.add_observer(self)
        logging.info(f"{self}当前定价方法为{engine.engine_type.value}")

    def remove_self(self):
        """删除对象自己，del自己之前，先将自己从被观察者的列表中移除"""
        self.engine.process.spot.remove_observer(self)
        del self

    def update(self, observable, *args, **kwargs):
        """收到标的价格变化的通知时，自动更新是否已经触碰障碍"""
        if global_evaluation_date() > self.stop_date:  # 如果估值日已经过了终止日，维持原来的状态
            return
        if global_auto_update_status() and observable == self.engine.process.spot:
            if self.obs_type == ExerciseType.American:
                # TODO: 目前美式观察障碍期权的update仅支持连续观察，或者每日观察（因为每周、每月观察日不确定，需要额外输入参数）
                if observable.data >= self.barrier and self.updown == UpDown.Up:
                    self.status = StatusType.UpTouch
                    if self.inout == InOut.Out and self.payment_type == PaymentType.Hit:
                        self.stop_date = global_evaluation_date()
                        logging.info(f"{self}【已向上触碰】，终止日为 {self.stop_date}")
                elif observable.data <= self.barrier and self.updown == UpDown.Down:
                    self.status = StatusType.DownTouch
                    if self.inout == InOut.Out and self.payment_type == PaymentType.Hit:
                        self.stop_date = global_evaluation_date()
                        logging.info(f"{self}【已向下触碰】，终止日为 {self.stop_date}")
            elif self.obs_type == ExerciseType.European:  # 欧式到期观察
                if global_evaluation_date() == self.end_date:  # 如果估值日是到期日
                    if observable.data >= self.barrier and self.updown == UpDown.Up:
                        self.status = StatusType.UpTouch
                    elif observable.data <= self.barrier and self.updown == UpDown.Down:
                        self.status = StatusType.DownTouch

    def __repr__(self):
        """返回期权的描述"""
        if self.obs_type == ExerciseType.American:
            if self.discrete_obs_interval is None:
                obs_type = "连续观察"
            else:
                obs_type = "离散观察"
        elif self.obs_type == ExerciseType.European:
            obs_type = "到期观察"
        else:
            raise ValueError(f"ExerciseType {self.obs_type} is not supported")
        return f"{obs_type}{self.barrier_type.value.text}期权"

    @property
    def callput(self):
        return self._callput

    @callput.setter
    def callput(self, value):
        self._callput = value
        self.status = StatusType.NoTouch  # 产品参数改变，重置是否已经敲出或敲入
        self.barrier_type = BarrierType.get_category(self._updown, self._inout, self._callput)

    @property
    def updown(self):
        return self._updown

    @updown.setter
    def updown(self, value):
        self._updown = value
        self.status = StatusType.NoTouch  # 产品参数改变，重置是否已经敲出或敲入
        self.barrier_type = BarrierType.get_category(self._updown, self._inout, self._callput)

    @property
    def inout(self):
        return self._inout

    @inout.setter
    def inout(self, value):
        self._inout = value
        self.status = StatusType.NoTouch  # 产品参数改变，重置是否已经敲出或敲入
        self.barrier_type = BarrierType.get_category(self._updown, self._inout, self._callput)

    @time_this
    def price(self, t: datetime.date = None, spot=None):
        """计算期权价格
        Args:
            t: datetime.date，计算期权价格的日期
            spot: float，标的价格
        Returns: 期权现值
        """
        self.validate_parameters(t=t)
        if self.status != StatusType.NoTouch:  # 如果已经触碰障碍，则敲出价值=现金返还的现值，敲入价值=香草期权
            calculate_date = global_evaluation_date() if t is None else t
            tenor = (self.end_date - self.start_date).days / self.annual_days.value  # 从起始日到终止日的自然日时间（年）
            _maturity = (self.end_date - calculate_date).days / self.annual_days.value
            # 敲出现金返还
            if self.inout == InOut.Out:  # 已经发生敲出，支付现金返还rebate
                if self.rebate_annual:
                    _rebate = self.rebate * tenor * self.s0  # 年化百分比（仅适用于到期支付rebate的情形，敲出立即支付需自行按敲出时间计算）
                else:
                    _rebate = self.rebate  # 绝对数值
                if self.payment_type == PaymentType.Hit:  # 立即支付
                    if self.rebate_annual:
                        elapsed_time = (calculate_date - self.start_date).days / self.annual_days.value
                        _rebate = self.rebate * self.s0 * elapsed_time
                    return _rebate
                elif self.payment_type == PaymentType.Expire:  # 到期支付
                    return _rebate * self.engine.process.interest.disc_factor(_maturity)
                else:
                    raise ValueError(f"未知的支付方式{self.payment_type}")
            elif self.inout == InOut.In:  # 已经发生敲入，返回香草期权的价值
                vanilla_engine = AnalyticVanillaEuEngine(stoch_process=self.engine.process)
                in_option = VanillaOption(exercise_type=ExerciseType.European, callput=self.callput, strike=self.strike,
                                          start_date=self.start_date, end_date=self.end_date, engine=vanilla_engine)
                # 保底收益率
                _floor_yield = 0
                if self.floor_yield != 0:  # 保底收益率需要输入期初价格s0作为名义本金
                    _floor_yield = self.floor_yield * tenor * self.s0
                return in_option.price(t, spot) + _floor_yield * self.engine.process.interest.disc_factor(_maturity)
            else:
                raise ValueError("inout只能是InOut.In或InOut.Out")
        else:  # 尚未触碰障碍，正常定价
            if self.engine.engine_type == EngineType.PdeEngine:  # 接口特殊是因为PDE引擎兼容了单双边障碍，所以需要把barrier填入bound
                if self.updown == UpDown.Up:
                    bound = (None, self.barrier)
                else:
                    bound = (self.barrier, None)
                price = self.engine.calc_present_value(prod=self, t=t, spot=spot, bound=bound,
                                                       rebate=(self.rebate, self.rebate))
            else:
                price = self.engine.calc_present_value(prod=self, t=t, spot=spot)
        return price



#!/user/bin/env python
# -*- coding: utf-8 -*-
"""
Copyright (C) 2024 Galaxy Technologies
Licensed under the Apache License, Version 2.0
"""
import gc
import datetime
from abc import ABCMeta
from contextlib import suppress
import importlib
from typing import List
import numpy as np
from scipy.stats import norm, qmc
from ..processes import StochProcessBase
from ..utilities.enums import RandsMethod, LdMethod, ProcessType, EngineType
from ..date_handler import global_evaluation_date
from ..utilities.utility import logging
from .engine_base import PricingEngineBase


class McEngine(PricingEngineBase, metaclass=ABCMeta):
    """蒙特卡洛模拟定价引擎基类
    观察者，观察随机过程process对象，当process对象的属性变化时，自动更新状态，会重新生成价格路径"""
    engine_type = EngineType.McEngine

    def __init__(self, stoch_process: StochProcessBase = None, n_path=100000, rands_method=RandsMethod.LowDiscrepancy,
                 antithetic_variate=True, ld_method=LdMethod.Sobol, seed=0, *, rands=None, s_paths=None, var_paths=None,
                 s=None, r=None, q=None, vol=None):
        """构造函数
        Args:
            stoch_process: 随机过程StochProcessBase对象
            n_path: int，MC模拟路径数
            rands_method: 生成随机数方法，RandsMethod枚举类，Pseudorandom伪随机数/LowDiscrepancy低差异序列
            antithetic_variate: bool，是否使用对立变量法
            ld_method: 若使用了低差异序列，指定低差异序列方法，LdMethod枚举类，Sobol序列/Halton序列
            seed: int，随机数种子
        在未设置stoch_process时，(stoch_process=None)，会默认创建BSMprocess，需要输入以下变量进行初始化
            s: float，标的价格
            r: float，无风险利率
            q: float，分红/融券率
            vol: float，波动率
        """
        super().__init__(stoch_process=stoch_process, s=s, r=r, q=q, vol=vol)
        # MC模拟路径数
        if n_path % 2 != 0 and antithetic_variate:
            n_path += 1
        self.n_path = n_path
        self.rands_method = rands_method  # 生成随机数方法
        self._ld_method = ld_method  # 低差异序列方法
        self._antithetic_variate = antithetic_variate  # 是否使用对立变量
        self.seed = seed  # 随机数种子
        # 如果传入了路径矩阵，则不再生成新的路径矩阵
        if rands is None:
            self.rands = None  # 随机数矩阵
            self.__reset_rands = True  # 重置随机数标志位
        else:
            self.rands = rands  # 随机数矩阵
            self.__reset_rands = False  # 重置随机数标志位
        if s_paths is None and var_paths is None:
            self.s_paths = None  # 价格路径矩阵
            self.var_paths = None  # 方差路径矩阵
            self.__reset_paths = True  # 重置路径标志位
        else:  # 用户只要传入了一个路径矩阵，就视为用户已经设置了路径矩阵，其他矩阵也不再重新生成
            self.s_paths = s_paths  # 价格路径矩阵
            self.var_paths = var_paths  # 方差路径矩阵
            self.__reset_paths = False  # 重置路径标志位
            self.__reset_rands = False  # 重置随机数标志位

    def set_stoch_process(self, stoch_process):
        """设置随机过程，先将自己从原来的随机过程的观察者列表中移除，再将自己加入新的随机过程的观察者列表"""
        with suppress(AttributeError, ValueError):
            self.process.remove_observer(self)
        self.process = stoch_process
        stoch_process.add_observer(self)  # 将本实例加入stoch_process观察者列表
        self.rands = None  # 更换随机过程之后需要重新生成随机数
        self.__reset_rands = True  # 更换随机过程之后需要重新生成随机数
        self.__reset_paths = True  # 更换随机过程之后需要重新生成路径

    def set_s_paths(self, s_paths):
        """设置价格路径矩阵"""
        self.s_paths = s_paths
        self.__reset_paths = False  # 用户设置了路径矩阵，不需要重新生成路径
        self.__reset_rands = False  # 重置随机数标志位

    def remove_self(self):
        """删除对象自己，del自己之前，先将自己从被观察者的列表中移除"""
        self.process.remove_observer(self)
        del self

    def update(self, observable, *args, **kwargs):
        """被观察者的值发生变化时，自动调用update方法，通知观察者"""
        self.__reset_paths = True  # process自动向观察者发送通知, 参数已改变, 重置路径标志位，重新生成路径

    def reset_paths_flag(self):
        """重置路径标志位，重新生成路径"""
        self.__reset_paths = True

    @property
    def antithetic_variate(self):
        return self._antithetic_variate

    @antithetic_variate.setter
    def antithetic_variate(self, new_value):
        self._antithetic_variate = new_value
        self.__reset_paths = True  # 重置路径标志位，重新生成路径
        self.__reset_rands = True  # 重置随机数标志位，重新生成随机数

    @property
    def ld_method(self):
        return self._ld_method

    @ld_method.setter
    def ld_method(self, new_value):
        self._ld_method = new_value
        self.__reset_paths = True  # 重置路径标志位，重新生成路径
        self.__reset_rands = True  # 重置随机数标志位，重新生成随机数

    def randoms_generator(self, shape):  # 参数rands_method控制随机数获取方法，LD_method低差异序列生成方法；
        """生成随机数，返回shape形状的随机数矩阵
        如果需要的随机数矩阵的形状大于已有随机数矩阵的形状，则重新生成随机数矩阵，否则复用已有随机数矩阵
            rands_method: 控制随机数生成方法，包括伪随机数和低差异序列两种方法
            _ld_method: 控制低差异序列生成方法，包括Sobol和Halton两种低差异序列类型
        Args:
            shape: np.ndarray，(n_row, n_col)，随机数矩阵的形状
        Returns: self.rands，shape形状的随机数矩阵
        """
        if (self.rands is not None and (not self.__reset_rands)
                and self.rands.shape[0] >= shape[0] and self.rands.shape[1] >= shape[1]):
            logging.info("复用已有随机数")
            return self.rands[:shape[0], :shape[1]]
        else:
            np.random.seed(self.seed)
            if self.rands_method == RandsMethod.Pseudorandom:
                self.rands = np.random.standard_normal(shape)
            elif self.rands_method == RandsMethod.LowDiscrepancy:
                if self._ld_method == LdMethod.Sobol:
                    sampler = qmc.Sobol(d=shape[0], scramble=True, seed=self.seed)
                elif self._ld_method == LdMethod.Halton:
                    sampler = qmc.Halton(d=shape[0], scramble=True, seed=self.seed)
                else:
                    raise ValueError(
                        f'随机数生成方法输入错误，应为（Halton, Sobol）二者之一， 当前输入为{RandsMethod}')
                sampler.fast_forward(30)  # 跳过前三十项
                uniform_rands = sampler.random(shape[1]).transpose()
                np.random.shuffle(uniform_rands)
                self.rands = norm.ppf(uniform_rands)
            else:
                raise ValueError(
                    f'随机数生成方法输入错误，应为（RandsMethod.LowDiscrepancy, RandsMethod.Pseudorandom）二者之一， 当前输入为{RandsMethod}')
            self.__reset_rands = False  # 重置随机数标志位
            return self.rands

    def path_generator(self, n_step, spot=None, t_step_per_year=243):
        """根据标的种类和标的参数，生成价格路径，返回价格路径矩阵
        如果标的参数不变，且需要的价格路径矩阵的形状 ≤ 已有价格路径矩阵的形状，则复用已有价格路径矩阵，否则重新生成价格路径矩阵
        Args:
            n_step: int，价格路径的时间步数
            spot: float，标的期初价格
            t_step_per_year: int，每年的时间步数
        Returns: self.s_paths，价格路径矩阵
        """
        if (self.s_paths is not None and (not self.__reset_paths)
                and self.s_paths.shape[0] >= n_step + 1 and self.s_paths.shape[1] >= self.n_path
                and spot == self.process.spot()):
            logging.info("复用已有价格路径")
            return self.s_paths[:n_step + 1, :self.n_path]
        # else:
        return self._regenerate_paths(n_step, self.n_path, spot, t_step_per_year)

    def _regenerate_paths(self, n_step, n_path, spot, t_step_per_year):
        """生成价格路径的执行函数
        Args:
            n_step: int，价格路径的时间步数
            n_path: int，模拟路径数量
            spot: float，标的期初价格
            t_step_per_year: int，每年的时间步数
        Returns:
        """
        # self.s_paths = np.empty(shape=(n_step + 1, n_path))
        # self.s_paths[0] = spot
        dt = 1 / t_step_per_year

        if self.process() == ProcessType.BSProcess1D:
            if self.antithetic_variate:
                if n_path % 2 != 0:
                    n_path += 1
                half_n_path = n_path // 2
                self.rands = self.randoms_generator(shape=(n_step, half_n_path))
                rand_s = np.concatenate((self.rands, -self.rands), axis=1)
            else:
                self.rands = self.randoms_generator(shape=(n_step, n_path))
                rand_s = self.rands.copy()

            self.s_paths = np.empty(shape=(n_step + 1, n_path))
            self.s_paths[0] = spot
            for step in range(1, n_step + 1):
                self.s_paths[step] = self.process.evolve(dt * step, self.s_paths[step - 1], dt, rand_s[step - 1])

        elif self.process() == ProcessType.Heston:
            self.var_paths = np.empty(shape=(n_step + 1, n_path))
            if self.antithetic_variate:
                self.rands = self.randoms_generator(shape=(n_step, n_path))
                rand_s = self.rands[:, :self.rands.shape[1] // 2].copy()
                rand_v = self.rands[:, self.rands.shape[1] // 2:].copy()
                rand_v = self.process.var_rho * rand_s + np.sqrt(1 - self.process.var_rho ** 2) * rand_v
                rand_s = np.concatenate((rand_s, -rand_s), axis=1)
                rand_v = np.concatenate((rand_v, -rand_v), axis=1)
            else:
                self.rands = self.randoms_generator(shape=(n_step, n_path * 2))
                rand_s = self.rands[:, :self.rands.shape[1] // 2].copy()
                rand_v = self.rands[:, self.rands.shape[1] // 2:].copy()
                rand_v = self.process.var_rho * rand_s + np.sqrt(1 - self.process.var_rho ** 2) * rand_v

            self.s_paths = np.empty(shape=(n_step + 1, n_path))
            self.s_paths[0] = spot
            self.var_paths[0] = self.process.v0

            for step in range(1, n_step + 1):
                [self.s_paths[step], self.var_paths[step]] = self.process.evolve(dt * step, [self.s_paths[step - 1],
                                                                                             self.var_paths[step - 1]],
                                                                                 dt,
                                                                                 [rand_s[step - 1], rand_v[step - 1]])
            self.var_paths = np.maximum(self.var_paths, 0)  # heston方差非负修正：部分截断形式
        else:
            raise ValueError(f'随机过程类型输入错误，应为（ProcessType.BSProcess1D, ProcessType.Heston）二者之一，'
                             f'当前输入为{self.process()}')

        self.__reset_paths = False  # 重置路径标志位
        return self.s_paths

    def generate_reusable_price_paths(self, n_step=800, s_step=0.01, vol_step=0.01, r_step=0.01):
        """生成可复用的价格路径列表，包含计算希腊值需要的参数经过微小扰动的模拟价格路径. 仅支持BSM模型常数波动率、常数分红率、常数无风险利率
        对于一个指定的估值日，每个标的资产需要生成一个复用价格路径列表，因为每个标的资产有自己的波动率、分红率，
        两个标的资产相同的期权，其标的价格、波动率、分红率应该是相同的。另外，同一个估值日，所有期权的无风险利率应该是统一的。
        Args:
            n_step: int，价格路径的时间步数，默认生成800个交易日，因为大部分衍生品的到期时间不超过3年.
            s_step: float，价格步长。由于蒙特卡洛模拟使用有限差分计算希腊值时，步长较大时delta抖动较大，步长较小时gamma抖动较大，默认取0.01
            vol_step: float，默认vega计算波动率上升1%的期权价值变化
            r_step: float，默认rho计算无风险利率上升1%的期权价值变化
        Returns:
            s_paths_list: List，复用价格路径列表，用于有限差分法计算希腊值，其顺序为：
                          [原始价格路径，s-Δs价格路径，s+Δs价格路径，vol+0.01价格路径，r+0.01价格路径， # 价格路径的数据类型都是np.ndarray
                          s的原始数值，r的原始数值，s的步长（百分比），r的步长（百分比）]  # 这一行变量的数据类型都是float
        """
        s = self.process.spot.data
        r = self.process.interest.data
        vol = self.process.vol.data
        # 生成复用的随机数矩阵
        reusable_rands = self.randoms_generator(shape=(n_step, self.n_path))
        # 1. 生成原始价格路径
        s_paths_list = [self.path_generator(n_step=n_step, spot=s)]
        # 2. 生成s-Δs价格路径
        s_paths_list.append(self.path_generator(n_step=n_step, spot=(1 - s_step) * s))
        # 3. 生成s+Δs价格路径
        s_paths_list.append(self.path_generator(n_step=n_step, spot=(1 + s_step) * s))
        # 4. 生成vol+0.01价格路径
        self.process.vol.data = vol + vol_step
        s_paths_list.append(self.path_generator(n_step=n_step, spot=s))
        self.process.vol.data = vol
        # 4. 生成r+0.01价格路径
        self.process.interest.data = r + r_step
        s_paths_list.append(self.path_generator(n_step=n_step, spot=s))
        self.process.interest.data = r
        s_paths_list += [s, r, s_step, r_step]
        return s_paths_list

    def pv_and_greeks(self, prod, spot, s_paths_list: List=None, parallel=False):
        """根据复用的模拟价格路径计算期权希腊值
        Args:
            prod: Product类，期权产品对象
            spot: float, 标的价格，无实际用处，仅为了保持api参数统一
            s_paths_list: List，复用价格路径列表，其顺序为：
                          [原始价格路径，s-Δs价格路径，s+Δs价格路径，vol+0.01价格路径，r+0.01价格路径，
                          s的原始数值，r的原始数值，s的步长（百分比），r的步长（百分比）]  # 这一行变量的数据类型都是float
                          当串行计算时，路径数组是np.ndarray；当ray并行时，路径数组是ray.put存储到共享内存后的ObjectID
            parallel: bool，是否使用ray并行计算，默认False
        Returns:
            Dict[str:float]: {'pv': pv, 'delta': delta, 'gamma': gamma, 'theta': theta, 'vega': vega, 'rho': rho}
        """
        if s_paths_list is None:
            raise ValueError("McEngine.pv_and_greeks()必须输入复用价格路径列表s_paths_list.")
        assert len(s_paths_list) == 9, "复用价格路径列表s_paths_list应为长度9的列表，请使用McEngine的generate_reusable_price_paths方法生成"
        spot = s_paths_list[-4]
        r = s_paths_list[-3]
        s_step = s_paths_list[-2]
        r_step = s_paths_list[-1]
        if parallel:  # ray并行计算
            ray = importlib.import_module('ray')
        # 1.估值列表，存储不同参数微扰的估值结果，最后用于有限差分法计算希腊值。
        v_list = []  # 用于存储每种价格路径的估值结果
        # 2. 依次计算 s-Δs价格路径，s+Δs价格路径，vol+0.01价格路径，r+0.01价格路径 的估值
        for i in range(5):
            if i == 4:  # 计算rho的微扰后估值时，需要将无风险利率+1%，因为估值过程中，除了价格路径，折现也要用到r
                self.process.interest.data = r + r_step
            if parallel:  # ray并行计算，需要及时释放内存
                tmp_path = ray.get(s_paths_list[i])
                self.set_s_paths(tmp_path)
                del tmp_path
                gc.collect()
            else:  # 串行计算
                self.set_s_paths(s_paths_list[i])
            v_list.append(prod.price())
            if i == 0:  # 原始路径，既可用于计算pv，也可计算theta
                # 计算theta，假设标的价格不变，交易日推后一天的期权价值变化
                next_date = prod.trade_calendar.advance(global_evaluation_date(), datetime.timedelta(days=1))
                theta = prod.price(t=next_date, spot=spot) - v_list[0]
            if i == 4:  # 还原原始r
                self.process.interest.data = r
        # 4. 有限差分法计算希腊值
        ds = spot * s_step
        result = {"pv": v_list[0],
                  "delta": (v_list[2] - v_list[1]) / (2 * ds),
                  "gamma": (v_list[2] - 2 * v_list[0] + v_list[1]) / ds ** 2,
                  "vega": (v_list[3] - v_list[0]),
                  "theta": theta,
                  "rho": (v_list[4] - v_list[0]),
                  }
        return result


class McReusablePaths(McEngine):
    """专门用于生成可复用的Monte Carlo模拟价格路径的类"""

    def calc_present_value(self, prod, t: datetime.date = None, spot: float = None):
        """计算期权的现值，本来是抽象基类McEngine的抽象方法，
        这里重写calc_present_value，使子类McReusablePathGenerator可以被实例化"""
        raise NotImplementedError(
            "McReusablePathGenerator未实现计算期权价值的功能，该类仅用于生成可复用的Monte Carlo模拟价格路径！")



#!/user/bin/env python
# -*- coding: utf-8 -*-
"""
Copyright (C) 2024 Galaxy Technologies
Licensed under the Apache License, Version 2.0
"""
from functools import partial
from contextlib import suppress
import datetime
import numpy as np
from pricelib.common.utilities.enums import CallPut, EngineType, StatusType
from pricelib.common.date_handler import AnnualDays, CN_CALENDAR, global_evaluation_date
from pricelib.common.utilities.patterns import Observer, global_auto_update_status
from pricelib.common.utilities.utility import time_this, logging
from pricelib.pricing_engines.integral_engines import QuadAutoCallEngine
from pricelib.pricing_engines.fdm_engines import FdmSnowBallEngine
from pricelib.pricing_engines.mc_engines import MCAutoCallEngine
from .autocallable_base import AutocallableBase


class AutoCall(AutocallableBase, Observer):
    """Autocall Note(二元小雪球)，不带敲入的自动赎回结构，每月观察敲出，到期未敲出获得红利票息(保本产品)"""

    def __init__(self, s0, *, barrier_out, coupon_out, coupon_div=0, callput=CallPut.Call, lock_term=1,
                 maturity=None, start_date=None, end_date=None, obs_dates=None, pay_dates=None, margin_lvl=1,
                 engine=None, trade_calendar=CN_CALENDAR, annual_days=AnnualDays.N365, t_step_per_year=243,
                 status=StatusType.NoTouch, stoch_process=None, s=None, r=None, q=None, vol=None, engine_method=None):
        """构造函数
        Args:
            s0: float，标的初始价格
            barrier_out: float，敲出障碍价，绝对值/百分比
            coupon_out: float，敲出票息，百分比，年化
            coupon_div: float，红利票息，百分比，年化
            callput: 看涨看跌，CallPut枚举类，默认为Call
            lock_term: int，锁定期，单位为月，锁定期内不触发敲出
            margin_lvl:  float，保证金比例，默认为1，即无杠杆
            status: 敲出状态，StatusType枚举类，默认为NoTouch, 即未敲出
        时间参数: 要么输入年化期限，要么输入起始日和到期日；敲出观察日和票息支付日可缺省
            maturity: float，年化期限
            start_date: datetime.date，起始日
            end_date: datetime.date，到期日
            obs_dates: List[datetime.date]，敲出观察日，可缺省，缺省时会自动生成每月观察的敲出日期序列(已根据节假日调整)
            pay_dates: List[datetime.date]，票息支付日，可缺省，长度需要与敲出观察日数一致，如不指定则默认为敲出观察日
            trade_calendar: 交易日历，Calendar类，默认为中国内地交易日历
            annual_days: int，每年的自然日数量
            t_step_per_year: int，每年的交易日数量
        可选参数:
            engine: 定价引擎，PricingEngine类
                    蒙特卡洛: MCAutoCallEngine
                    PDE: FdmSnowBallEngine
                    积分法: QuadAutoCallEngine
            engine_method: str, 定价引擎种类。若未输入engine，可以用engine_method快捷生成指定定价引擎。可选MC/MonteCarlo、PDE/FDM、QUAD。
            若未提供引擎engine、也未指定引擎种类engine_method的情况下，如果提供了 随机过程 或 标的价格、无风险利率、分红/融券率、波动率，
            则默认使用 积分法 定价引擎 QuadAutoCallEngine
            stoch_process: 随机过程StochProcessBase对象
            s: float，标的价格
            r: float，无风险利率
            q: float，分红/融券率
            vol: float，波动率
        """
        super().__init__(s0=s0, maturity=maturity, start_date=start_date, end_date=end_date, lock_term=lock_term,
                         trade_calendar=trade_calendar, obs_dates=obs_dates, pay_dates=pay_dates, margin_lvl=margin_lvl,
                         t_step_per_year=t_step_per_year, annual_days=annual_days, status=status)
        len_obs_dates = len(self.obs_dates.date_schedule)
        self.barrier_out = np.ones(len_obs_dates) * barrier_out
        self.coupon_out = np.ones(len_obs_dates) * coupon_out
        self.coupon_div = coupon_div
        self.barrier_in = np.zeros(len_obs_dates)  # 小雪球无敲入
        self.callput = callput
        self.margin_lvl = margin_lvl  # 预付金比例
        self.stop_date = self.end_date  # 停止日期，默认等于到期日。发生敲出事件，停止日期更新为敲出日期
        if engine is not None:
            self.set_pricing_engine(engine)
        else:
            if engine_method is None:
                default_engine = QuadAutoCallEngine  # 默认QUAD定价引擎
            elif engine_method in ["PDE", "FDM"]:
                default_engine = FdmSnowBallEngine
            elif engine_method in ["MC", "MonteCarlo"]:
                default_engine = MCAutoCallEngine
            elif engine_method == "QUAD":
                default_engine = QuadAutoCallEngine
            else:
                raise NotImplementedError(f"不支持的engine_method{engine_method}, {self.__repr__()}仅支持MC/PDE/QUAD")
            if stoch_process is not None:
                self.set_pricing_engine(default_engine(stoch_process))
            elif s is not None and r is not None and q is not None and vol is not None:
                self.set_pricing_engine(default_engine(s=s, r=r, q=q, vol=vol))

    def set_pricing_engine(self, engine):
        """设置定价引擎"""
        with suppress(AttributeError, ValueError):
            self.engine.process.spot.remove_observer(self)
        self.engine = engine
        self.engine.process.spot.add_observer(self)
        logging.info(f"{self}当前定价方法为{engine.engine_type.value}")

    def remove_self(self):
        """删除对象自己，del自己之前，先将自己从被观察者的列表中移除"""
        self.engine.process.spot.remove_observer(self)
        del self

    def update(self, observable, *args, **kwargs):
        """收到标的价格变化的通知时，自动更新自动赎回结构是否已经敲入或敲出"""
        if global_auto_update_status():
            calculate_date = global_evaluation_date()
            if calculate_date < self.start_date or calculate_date > self.stop_date:
                return  # 如果在起始日前或终止日之后，状态维持不变
            if observable == self.engine.process.spot:
                if self.status == StatusType.UpTouch:  # 如果已敲出，维持原来的状态
                    return
                if calculate_date in self.obs_dates.date_schedule:  # 如果估值日是敲出观察日
                    obs_idx = self.obs_dates.date_schedule.index(calculate_date)
                    if observable.data >= self.barrier_out[obs_idx]:  # 如果标的价格大于等于敲出价，敲出
                        self.status = StatusType.UpTouch
                        self.stop_date = global_evaluation_date()
                        return


    def __repr__(self):
        """返回期权的描述"""
        return f"Auto{self.callput.name} Note(二元小雪球)"

    @time_this
    def price(self, t: datetime.date = None, spot=None):
        """计算期权价格
        Args:
            t: datetime.date，计算期权价格的日期
            spot: float，标的价格
        Returns: 期权现值
        """
        self.validate_parameters(t=t)
        if self.engine.engine_type == EngineType.PdeEngine and self.callput == CallPut.Put:
            logging.warning("PDE引擎暂不支持AutoPut定价")
            return np.nan
        else:
            price = self.engine.calc_present_value(prod=self, t=t, spot=spot)
            return price

    def delta(self, spot=None, step=None, *args, **kwargs):
        """求价格spot处的delta值
        会先尝试调用定价引擎的delta方法，如果没有则使用通用的差分法计算
        Args:
            spot: float，价格的绝对值
            step: float，微扰步长，默认为价格spot的1%
        Returns: delta = ∂V/∂S
        """
        try:
            if self.engine.engine_type != EngineType.PdeEngine or self.callput != CallPut.Put:  # PDE引擎暂不支持AutoPut定价
                return self.engine.delta(prod=self, spot=spot, step=step, *args, **kwargs)
        except:
            self.validate_parameters()
            spot = self.engine.process.spot() if spot is None else spot
            step = spot * 0.001 if step is None else step
            up_price = self.price(spot=spot + step)
            down_price = self.price(spot=spot - step)
            delta = (up_price - down_price) / (2 * step)
            return delta

    def gamma(self, spot=None, step=None, *args, **kwargs):
        """求价格spot处的gamma值
        会先尝试调用定价引擎的gamma方法，如果没有则使用通用的差分法计算
        Args:
            spot: float，价格的绝对值
            step: float，微扰步长，默认为价格spot的1%
        Returns: gamma = ∂2V/∂S2
        """
        try:
            if self.engine.engine_type != EngineType.PdeEngine or self.callput != CallPut.Put:  # PDE引擎暂不支持AutoPut定价
                return self.engine.gamma(prod=self, spot=spot, step=step, *args, **kwargs)
        except:
            self.validate_parameters()
            spot = self.engine.process.spot() if spot is None else spot
            step = spot * 0.01 if step is None else step
            up_price = self.price(spot=spot + step)
            down_price = self.price(spot=spot - step)
            mid_price = self.price(spot=spot)
            gamma = (up_price - 2 * mid_price + down_price) / (step ** 2)
            return gamma

    def vega(self, spot=None, step=0.01, *args, **kwargs):
        """计算波动率上升1%的vega todo:目前只支持常数波动率
        会先尝试调用定价引擎的vega方法，如果没有则使用通用的差分法计算
        Args:
            t: int，期权价值矩阵的列(时间)的索引
            spot: float，价格的绝对值
            step: float，微扰步长，默认为1%
        Returns: vega = ∂V/∂σ
        """
        try:
            if self.engine.engine_type != EngineType.PdeEngine or self.callput != CallPut.Put:  # PDE引擎暂不支持AutoPut定价
                return self.engine.vega(prod=self, spot=spot, step=step, *args, **kwargs)
        except:
            self.validate_parameters()
            spot = self.engine.process.spot() if spot is None else spot
            last_vol = self.engine.process.vol.data
            last_price = self.price(spot=spot)
            self.engine.process.vol.data = last_vol + step
            new_price = self.price(spot=spot)
            self.engine.process.vol.data = last_vol
            vega = (new_price - last_price)
            return vega

    def theta(self, spot=None, step=1, *args, **kwargs):
        """计算一天的theta值
        会先尝试调用定价引擎的theta方法，如果没有则使用通用的差分法计算
        Args:
            t: int，期权价值矩阵的列(时间)的索引
            spot: float，价格的绝对值
            step: int，时间步长，默认为1天
        Returns: theta = ∂V/∂t
        """
        try:
            if self.engine.engine_type != EngineType.PdeEngine or self.callput != CallPut.Put:  # PDE引擎暂不支持AutoPut定价
                return self.engine.theta(prod=self, spot=spot, step=step, *args, **kwargs)
        except:
            try:
                spot = self.engine.process.spot() if spot is None else spot
                step = datetime.timedelta(days=step)
                origin_price = self.price(spot=spot)
                next_date = self.trade_calendar.advance(global_evaluation_date(), step)
                new_price = self.price(t=next_date, spot=spot)
                theta = (new_price - origin_price)
                return theta
            except:  # 可能遇到到期日估值，无法计算theta
                calculate_date = global_evaluation_date()
                if calculate_date > self.trade_calendar.advance(self.end_date, datetime.timedelta(days=1)):
                    raise ValueError(f"估值日必须小于等于到期日, 当前估值日为{calculate_date}, 到期日为{self.end_date}.")
                return np.nan  # 到期日估值，无法计算theta,返回nan

    def rho(self, spot=None, step=0.01, *args, **kwargs):
        """计算无风险利率上升1%的rho todo:目前只支持常数无风险利率
        会先尝试调用定价引擎的rho方法，如果没有则使用通用的差分法计算
        Args:
            t: int，期权价值矩阵的列(时间)的索引
            spot: float，价格的绝对值
            step: float，微扰步长，默认为1%
        Returns: rho = ∂V/∂r
        """
        try:
            if self.engine.engine_type != EngineType.PdeEngine or self.callput != CallPut.Put:  # PDE引擎暂不支持AutoPut定价
                return self.engine.rho(prod=self, spot=spot, step=step, *args, **kwargs)
        except:
            self.validate_parameters()
            spot = self.engine.process.spot() if spot is None else spot
            last_r = self.engine.process.interest.data
            last_price = self.price(spot=spot)
            self.engine.process.interest.data = last_r + step
            new_price = self.price(spot=spot)
            self.engine.process.interest.data = last_r
            rho = (new_price - last_price)
            return rho

    def pv_and_greeks(self, spot=None, *args, **kwargs):
        """当一次性计算pv和5个greeks时，可以调用此函数，减少计算量
        会先尝试调用定价引擎的pv_and_greeks方法，如果没有则使用通用的差分法计算
        Args:
            t: int，期权价值矩阵的列(时间)的索引
            spot: float，价格的绝对值
        Returns:
            Dict[str:float]: {'pv': pv, 'delta': delta, 'gamma': gamma, 'theta': theta, 'vega': vega, 'rho': rho}
        """
        if self.engine.engine_type != EngineType.PdeEngine or self.callput != CallPut.Put:  # PDE引擎暂不支持AutoPut定价
            if hasattr(self.engine, "pv_and_greeks"):
                with suppress(AttributeError, ValueError):
                    self.price()
                    return self.engine.pv_and_greeks(prod=self, spot=spot, *args, **kwargs)
        self.validate_parameters()
        spot = self.engine.process.spot() if spot is None else spot
        pv = np.float64(self.price(spot=spot))
        s_step = spot * 0.01
        up_price = self.price(spot=spot + s_step)
        down_price = self.price(spot=spot - s_step)
        delta = (up_price - down_price) / (2 * s_step)
        gamma = (up_price - 2 * pv + down_price) / (s_step ** 2)

        last_vol = self.engine.process.vol.data
        self.engine.process.vol.data = last_vol + 0.01
        new_price = self.price(spot=spot)
        self.engine.process.vol.data = last_vol
        vega = (new_price - pv)

        last_r = self.engine.process.interest.data
        self.engine.process.interest.data = last_r + 0.01
        new_price = self.price(spot=spot)
        self.engine.process.interest.data = last_r
        rho = (new_price - pv)

        try:  # 可能遇到到期日估值，无法计算theta
            t_step = datetime.timedelta(days=1)
            next_date = self.trade_calendar.advance(global_evaluation_date(), t_step)
            new_price = self.price(t=next_date, spot=spot)
            theta = (new_price - pv)
        except:
            theta = np.nan
        return {'pv': pv, 'delta': delta, 'gamma': gamma, 'vega': vega, 'theta': theta, 'rho': rho}



#!/user/bin/env python
# -*- coding: utf-8 -*-
"""
Copyright (C) 2024 Galaxy Technologies
Licensed under the Apache License, Version 2.0
"""
import math
import numpy as np
from pricelib.common.utilities.enums import ExerciseType, PaymentType
from pricelib.common.date_handler import global_evaluation_date
from pricelib.common.pricing_engine_base import McEngine


class MCDoubleSharkEngine(McEngine):
    """双鲨期权 Monte Carlo 模拟定价引擎"""

    def calc_present_value(self, prod, t=None, spot=None):
        """计算现值
        Args:
            prod: Product产品对象
            t: datetime.date，估值日; 如果是None，则使用全局估值日globalEvaluationDate
            spot: float，估值日标的价格，如果是None，则使用随机过程的当前价格
        Returns: float，现值
        """
        if spot is None:
            spot = self.process.spot()
        calculate_date = global_evaluation_date() if t is None else t

        _maturity = (prod.end_date - calculate_date).days / prod.annual_days.value
        _maturity_business_days = prod.trade_calendar.business_days_between(calculate_date, prod.end_date)

        _elapsed_days = prod.trade_calendar.business_days_between(prod.start_date, calculate_date)  # 从起始日到估值日的交易日时间（天）
        _tenor = (prod.end_date - prod.start_date).days / prod.annual_days.value  # 从起始日到到期日的自然日时间（年）
        _floor_yield = self._preprocess_floor_yield(prod, _tenor)

        r = self.process.interest(_maturity)

        # 当估值日就是终止日时，直接用spot计算payoff，无需随机数矩阵
        if (calculate_date == prod.end_date) and isinstance(spot, np.ndarray):
            # 如果spot是一个np.ndarray一维价格向量，返回盈亏测算向量，即每个价格对应的现值，而不是平均值
            _tenor = (prod.end_date - prod.start_date).days / prod.annual_days.value  # 从起始日到终止日的自然日时间（年）
            _floor_yield = self._preprocess_floor_yield(prod, _tenor)
            value = self.calc_maturity_payoff(prod, np.array([spot, ]), _tenor, _floor_yield)
            return value
        elif calculate_date == prod.stop_date:  # spot是一个数值，计算到期日payoff
            _tenor = _tenor = (prod.stop_date - prod.start_date).days / prod.annual_days.value  # 从起始日到终止日的自然日时间（年）
            _floor_yield = self._preprocess_floor_yield(prod, _tenor)
            value = self.calc_maturity_payoff(prod, np.array([spot, ]), _tenor, _floor_yield)[0]
            return value
        else:
            if spot != self.process.spot():
                self.reset_paths_flag()  # 重置路径标志位，重新生成路径
            paths = self.path_generator(n_step=_maturity_business_days, spot=spot,
                                        t_step_per_year=prod.t_step_per_year).copy()

        call_payoff = (paths[-1] - prod.strike[1]) * prod.parti[1]
        put_payoff = (prod.strike[0] - paths[-1]) * prod.parti[0]
        # 美式：
        if prod.obs_type == ExerciseType.American:
            if prod.discrete_obs_interval is None:  # 每日观察
                obs_points = np.arange(0, _maturity_business_days + 1, 1)
            else:  # 均匀离散观察
                dt_step = prod.discrete_obs_interval * prod.t_step_per_year
                obs_points = np.flip(np.round(np.arange(_maturity_business_days, 0, -dt_step)).astype(int))
                obs_points = np.concatenate((np.array([0]), obs_points))
                obs_points[obs_points > _maturity_business_days] = _maturity_business_days  # 防止闰年导致的下标越界

            # 双鲨结构是双边敲出
            step_index = np.tile(obs_points, (self.n_path, 1)).T
            hit_lower = np.min(
                np.where(paths[obs_points] <= prod.bound[0], step_index.astype(int), _maturity_business_days + 1),
                axis=0)
            hit_upper = np.min(
                np.where(paths[obs_points] >= prod.bound[1], step_index.astype(int), _maturity_business_days + 1),
                axis=0)
            if prod.payment_type == PaymentType.Hit:  # 发生敲出时，立即支付现金返还
                _rebate_lower = self._preprocess_extra_returns(prod,
                                                               (_elapsed_days + hit_lower[hit_lower < hit_upper]
                                                                ) / prod.t_step_per_year, [prod.rebate[0], ])
                _rebate_upper = self._preprocess_extra_returns(prod,
                                                               (_elapsed_days + hit_upper[hit_lower > hit_upper]
                                                                ) / prod.t_step_per_year, [prod.rebate[1], ])
                value = np.mean(
                    np.where(hit_lower < hit_upper, _rebate_lower * np.exp(-r * hit_lower / prod.t_step_per_year),
                             np.where(hit_lower > hit_upper,
                                      _rebate_upper * np.exp(-r * hit_upper / prod.t_step_per_year),
                                      (np.where(call_payoff > 0, call_payoff, np.where(put_payoff > 0, put_payoff, 0))
                                       + _floor_yield) * math.exp(-r * _maturity))))
            elif prod.payment_type == PaymentType.Expire:
                _rebate = self._preprocess_extra_returns(prod, _tenor, prod.rebate)
                value = np.mean(np.where(hit_lower < hit_upper, _rebate[0],
                                         np.where(hit_lower > hit_upper, _rebate[1],
                                                  (np.where(call_payoff > 0, call_payoff,
                                                            np.where(put_payoff > 0, put_payoff, 0)) + _floor_yield
                                                   )))) * math.exp(-r * _maturity)
            else:
                raise ValueError("PaymentType must be Hit or Expire")

        # 欧式：
        elif prod.obs_type == ExerciseType.European:
            # 双鲨结构是双边敲出
            value = self.calc_maturity_payoff(prod, paths[-1], _tenor, _floor_yield)[0]
            value = np.mean(value) * math.exp(-r * _maturity)
        else:
            raise ValueError("ExerciseType must be American or European")
        return value

    def calc_maturity_payoff(self, prod, s_vec, _tenor, _floor_yield=0.):
        """初始化终止时的期权价值，在障碍价格内侧，默认设定未敲入，未敲出
        Args:
            prod: Product产品对象
            s_vec: np.ndarray, 网格的价格格点
            _tenor: float, 从起始日到终止日的自然日时间（年）
            _floor_yield: float，处理后的保底收益绝对数值
        Returns:
            value: np.ndarray, 期末价值向量
        """
        call_payoff = (s_vec - prod.strike[1]) * prod.parti[1]
        put_payoff = (prod.strike[0] - s_vec) * prod.parti[0]
        _rebate = self._preprocess_extra_returns(prod, _tenor, prod.rebate)
        value = np.where(s_vec <= prod.bound[0], _rebate[0],
                         np.where(s_vec >= prod.bound[1], _rebate[1],
                                  np.where(call_payoff > 0, call_payoff,
                                           np.where(put_payoff > 0, put_payoff, 0)) + _floor_yield))
        return value

    @staticmethod
    def _preprocess_extra_returns(prod, period, rebate_list):
        """预处理可能存在的年化现金返还rebate
        Args:
            prod: Product产品对象
            period: float 或 np.ndarray, 从起始到期权终止的时间（年）
                    对欧式障碍、美式敲入而言是整个存续期；对美式敲出而言，期权可能会因为敲出提前终止，此时period是起始日到敲出日的时间
            rebate_list: List[float], 现金返还率或绝对数值
        Returns:
            _rebate: float，处理后的现金返还绝对数值
        """
        if prod.rebate_annual:  # 年化百分比
            _rebate = []
            for r in rebate_list:
                _rebate.append(r * period * prod.s0)
        else:
            _rebate = rebate_list  # 绝对数值
        return _rebate

    @staticmethod
    def _preprocess_floor_yield(prod, period):
        """预处理可能存在的年化保底收益floor_yield
        Args:
            prod: Product产品对象
            period: float 或 np.ndarray, 从起始到期权终止的时间（年）
                    对欧式障碍、美式敲入而言是整个存续期；对美式敲出而言，期权可能会因为敲出提前终止，此时period是起始日到敲出日的时间
        Returns:
            _floor_yield: float，处理后的保底收益绝对数值
        """
        if prod.floor_yield != 0:  # 保底收益率
            _floor_yield = prod.floor_yield * period * prod.s0
        else:
            _floor_yield = 0.0
        return _floor_yield



#!/user/bin/env python
# -*- coding: utf-8 -*-
"""
Copyright (C) 2024 Galaxy Technologies
Licensed under the Apache License, Version 2.0
"""
import datetime
from pricelib import (global_evaluation_date, set_evaluation_date, CN_CALENDAR, Schedule, BusinessConvention,
                      MonthAdjustment, EndOfMonthModify)


def test_calendar_schedule():
    set_evaluation_date(datetime.date(2022, 1, 5))
    assert global_evaluation_date() == datetime.date(2022, 1, 5)
    start_date = datetime.date(2022, 1, 5)
    end_date = start_date + datetime.timedelta(days=365) * 1
    assert end_date == datetime.date(2023, 1, 5)
    assert CN_CALENDAR.business_days_between(start_date, end_date) == 243

    obs_dates = Schedule(trade_calendar=CN_CALENDAR, start=start_date, end=end_date, freq='m',
                         lock_term=3, convention=BusinessConvention.FOLLOWING, correction=MonthAdjustment.YES,
                         endofmonthmodify=EndOfMonthModify.YES)
    assert obs_dates.date_schedule == [datetime.date(2022, 4, 6), datetime.date(2022, 5, 5),
                                       datetime.date(2022, 6, 6), datetime.date(2022, 7, 5),
                                       datetime.date(2022, 8, 5), datetime.date(2022, 9, 5),
                                       datetime.date(2022, 10, 10), datetime.date(2022, 11, 7),
                                       datetime.date(2022, 12, 5), datetime.date(2023, 1, 5)]
    assert obs_dates.count_business_days(start_date) == [58, 76, 97, 118, 141, 162, 181, 201, 221, 243]
    assert obs_dates.count_calendar_days(start_date) == [91, 120, 152, 181, 212, 243, 278, 306, 334, 365]



#!/user/bin/env python
# -*- coding: utf-8 -*-
"""
Copyright (C) 2024 Galaxy Technologies
Licensed under the Apache License, Version 2.0
"""
from abc import ABCMeta
import numpy as np
import numba as nb
from scipy.stats import norm
from ..utilities.enums import EngineType
from .engine_base import PricingEngineBase


class AnalyticEngine(PricingEngineBase, metaclass=ABCMeta):
    """解析定价引擎基类"""
    engine_type = EngineType.AnEngine


@nb.njit(cache=True, fastmath=True)
def bs_d1(s, k, t, r, sigma, q=0):
    """欧式股票期权BSM公式中的d1
    Args:
        s: float，标的价格
        k: float，执行价
        t: float，到期时间（年）
        r: float，无风险利率（常数）
        sigma: float，波动率（常数）
        q: float，连续分红率（股息率）（常数）
    Returns:
    """
    d1 = (np.log(s / k) + (r - q + 0.5 * sigma ** 2) * t) / (sigma * np.sqrt(t))
    return d1


@nb.njit(cache=True, fastmath=True)
def bs_d2(s, k, t, r, sigma, q=0):
    """欧式股票期权BSM公式中的d2
    Args:
        s: float，标的价格
        k: float，执行价
        t: float，到期时间（年）
        r: float，无风险利率（常数）
        sigma: float，波动率（常数）
        q: float，连续分红率（股息率）（常数）
    Returns:
    """
    d2 = (np.log(s / k) + (r - q - 0.5 * sigma ** 2) * t) / (sigma * np.sqrt(t))
    return d2


def bs_formula(s, k, t, r, sigma, sign=1, q=0):
    """欧式股票期权Black-Scholes-Merton公式, 针对现货（股票）期权的定价模型。认购期权的sign为1, 认沽期权的sign为-1
    若股票有稳定的连续分红率q，请传入参数q。股票在0时刻价格 = S(0)exp(qT)，股票在T时间中的分红 = S(0) * [exp(qT) - 1]
    Args:
        s: float，标的价格
        k: float，执行价
        t: float，到期时间（年）
        r: float，无风险利率（常数）
        sigma: float，波动率（常数）
        sign: int，期权方向，1为认购期权，-1为认沽期权
        q: float，连续分红率（股息率）（常数）
    Returns:
        float，欧式香草股票期权期权价格
    """
    if t == 0:  # 如果估值日就是到期日，直接返回payoff
        return max(sign * (s - k), 0)
    d1 = bs_d1(s=s, k=k, t=t, r=r, q=q, sigma=sigma)
    d2 = d1 - sigma * np.sqrt(t)
    return sign * np.exp(-q * t) * s * norm.cdf(sign * d1) - sign * np.exp(-r * t) * k * norm.cdf(sign * d2)

def black76_formula(f, k, t, r, sigma, sign=1, q=None):
    """期货期权Black76公式, 针对远期或期货合约的期权的定价模型。认购期权的sign为1, 认沽期权的sign为-1。
    期货期权的q通常被假设等于 r, 即持有标的资产的机会成本 b = r-q = 0, 原因是期货合约的持有成本与现货的持有成本相比可以忽略不计，
    因为期货合约是在未来某个时间点交割的标准化合约，并且通常不涉及实际持有标的资产，不需要支付像现货那样的存储、保险和融资成本。
    实践中，期货交易是保证金交易，不需要持有整个标的资产，相关的持有成本可以忽略，b 被设为零可以认为是出于简化模型的考虑。
    Args:
        f: float，标的远期或期货合约价格
        k: float，执行价
        t: float，到期时间（年）
        r: float，无风险利率（常数）
        sigma: float，波动率（常数）
        sign: int，期权方向，1为认购期权，-1为认沽期权
        q: float，固定等于r。该参数无作用，仅用于保持与bs_formula函数接口一致
    Returns:
        float，标的远期或期货合约的欧式香草期权价格
    """
    if t == 0:  # 如果估值日就是到期日，直接返回payoff
        return max(sign * (f - k), 0)
    d1 = bs_d1(s=f, k=k, t=t, r=r, q=r, sigma=sigma)
    d2 = d1 - sigma * np.sqrt(t)
    return sign * np.exp(-r * t) * (f * norm.cdf(sign * d1) - k * norm.cdf(sign * d2))


def euro_cash_or_nothing(s, k, t, r, sigma, sign=1, q=0):
    """欧式香草期权中的现金或无的价值，其中的现金是1(元)。主要是BSM公式中的一部分 N(d2)，这是在风险中性测度下，期权被执行的概率。
        例如看涨的现金1或无期权，到期时刻，St >= strike, payoff = 1(元) ; St < strike, payoff = 0.
        N(d2)积分推导BSM时，当 St > k，可推出积分变量 x>-d2。x是标准正态分布，所以 St > k 的可能性就是 N(d2).
    Args:
        s: float，标的价格
        k: float，执行价
        t: float，到期时间（年）
        r: float，无风险利率（常数）
        sigma: float，波动率（常数）
        sign: int，期权方向，1为认购期权，-1为认沽期权
        q: float，连续分红率（股息率）（常数）
    Returns:
        欧式香草期权中的现金1或无的价值
    """
    d2 = bs_d2(s=s, k=k, t=t, r=r, q=q, sigma=sigma)
    return np.exp(-r * t) * norm.cdf(sign * d2)


def euro_asset_or_nothing(s, k, t, r, sigma, sign=1, q=0):
    """欧式香草期权中的资产或无的价值。主要是BSM公式中的一部分 s*N(d1)，代表着行权得到股票的期望收益。
        例如看涨的资产或无期权，到期时刻，St > strike, payoff = St; St < strike, payoff = 0.
    Args:
        s: float，标的价格
        k: float，执行价
        t: float，到期时间（年）
        r: float，无风险利率（常数）
        sigma: float，波动率（常数）
        sign: int，期权方向，1为认购期权，-1为认沽期权
        q: float，连续分红率（股息率）（常数）
    Returns:
        欧式香草期权中的资产或无的价值
    """
    d1 = bs_d1(s=s, k=k, t=t, r=r, q=q, sigma=sigma)
    return np.exp(-q * t) * s * norm.cdf(sign * d1)



#!/user/bin/env python
# -*- coding: utf-8 -*-
"""
Copyright (C) 2024 Galaxy Technologies
Licensed under the Apache License, Version 2.0
"""
import math
import datetime
from contextlib import suppress
from pricelib.common.product_base.option_base import OptionBase
from pricelib.common.utilities.enums import CallPut, ExerciseType, AverageMethod, AsianAveSubstitution
from pricelib.common.date_handler import CN_CALENDAR, AnnualDays, global_evaluation_date
from pricelib.common.utilities.patterns import Observer, global_auto_update_status
from pricelib.common.utilities.utility import time_this, logging
from pricelib.pricing_engines.analytic_engines import AnalyticAsianEngine
from pricelib.pricing_engines.mc_engines import MCAsianEngine
from pricelib.pricing_engines.tree_engines import BiTreeAsianEngine


class AsianOption(OptionBase, Observer):
    """
    亚式期权：几何平均/算术平均；替代标的资产价格/替代执行价；增强亚式
        解析解只支持平均结算价，无增强亚式
        蒙特卡洛支持几何平均/算术平均，支持增强亚式，支持替代标的资产价格/替代执行价
        二叉树支持几何平均/算术平均，支持增强亚式，支持替代标的资产价格，不支持替代执行价
        todo:若定价时已经处于均价观察期内，需要获得当前已经实现的均价，并对行权价进行调整
    """
    exercise_type = ExerciseType.Asian

    def __init__(self, callput=CallPut.Call, ave_method=AverageMethod.Geometric, strike=None,
                 substitute=AsianAveSubstitution.Underlying, parti=1, enhanced=False, limited_price=None,
                 engine=None, maturity=None, start_date=None, end_date=None, obs_start=None, obs_end=None,
                 status=None, trade_calendar=CN_CALENDAR, annual_days=AnnualDays.N365, t_step_per_year=243,
                 stoch_process=None, s=None, r=None, q=None, vol=None, engine_method=None):
        """亚式期权初始化
        产品参数:
            callput: 看涨看跌类型，CallPut枚举类，Call/Put
            ave_method: 平均价计算方法，AverageMethod枚举类，几何平均Geometric/算术平均Arithmetic
            strike: float, 执行价
            substitute: 以均价代替标的资产价格/替代行权价，AsianAveSubstitution枚举类，Underlying/Strike
            parti: float, 参与率，默认为1
            enhanced: bool，是否为增强亚式期权，默认为False
            limited_price: float，增强型亚式限定标的资产价格的上/下限
            status: [int, float]，算术平均：[已经过的观察日天数, 已实现的采集价的累加值].未进入平均观察期时，默认为 [0, 0]
                                  几何平均：[已经过的观察日天数, 已实现的采集价的自然对数累加值].未进入平均观察期时，默认为 [0, 0]
        时间参数: 要么输入年化期限，要么输入起始日和到期日
            maturity: float，年化期限
            start_date: datetime.date，起始日
            end_date: datetime.date，到期日
            obs_start: datetime.date，观察期起始日
            obs_end: datetime.date，观察期终止日
            trade_calendar: 交易日历，Calendar类，默认为中国内地交易日历
            annual_days: int，每年的自然日数量
            t_step_per_year: int，每年的交易日数量
        可选参数:
            engine: 定价引擎，PricingEngine类
                    解析解: AnalyticAsianEngine
                           几何平均替代结算价亚式: Kemma Vorst (1990) 连续观察，平均价观察期为整个期权有效期内
                           算术平均替代结算价亚式: Turnbull Wakeman (1991) 近似估计，平均观察期的起点任意，终点为期权到期日
                    二叉树: BiTreeAsianEngine 只支持算术平均替代结算价亚式
                    蒙特卡洛: MCAsianEngine
                           几何平均/算术平均替代结算价亚式
                           几何平均/算术平均替代执行价亚式
                           增强亚式
            engine_method: str, 定价引擎种类。若未输入engine，可以用engine_method快捷生成指定定价引擎。可选Analytic、MC/MonteCarlo、PDE/FDM、QUAD。
            若未提供引擎engine、也未指定引擎种类engine_method的情况下，如果提供了 随机过程 或 标的价格、无风险利率、分红/融券率、波动率，
            则平均替代收盘价亚式默认使用解析解定价引擎；平均替代执行价亚式和增强亚式默认使用蒙特卡洛定价引擎
            stoch_process: 随机过程StochProcessBase对象
            s: float，标的价格
            r: float，无风险利率
            q: float，分红/融券率
            vol: float，波动率
        """
        super().__init__()
        self.trade_calendar = trade_calendar
        self.annual_days = annual_days
        self.t_step_per_year = t_step_per_year
        assert maturity is not None or (start_date is not None and end_date is not None), "Error: 到期时间或起止时间必须输入一个"
        self.start_date = start_date if start_date is not None else global_evaluation_date()  # 起始时间
        self.end_date = end_date if end_date is not None else (
                self.start_date + datetime.timedelta(days=round(maturity * annual_days.value)))  # 结束时间
        if self.trade_calendar.isholiday(self.end_date):  # 如果到期日不是交易日，向后顺延一个交易日
            self.end_date = self.trade_calendar.advance(self.end_date, datetime.timedelta(days=1))
        self.maturity = maturity if maturity is not None else (self.end_date - self.start_date).days / annual_days.value
        self.obs_start = obs_start if obs_start is not None else self.start_date
        self.obs_end = obs_end if obs_end is not None else self.end_date
        self.callput = callput
        self.ave_method = ave_method
        self.substitute = substitute
        self.parti = parti
        self.strike = strike
        self.enhanced = enhanced
        if enhanced and limited_price is None:
            # 增强型亚式的限定标的价格上/下限参数默认为执行价格
            self.limited_price = self.strike
        else:
            self.limited_price = limited_price
        assert strike is not None or substitute == AsianAveSubstitution.Strike, "Error: 替代平均结算价，必需输入执行价参数"
        if status is None:
            # 算术平均：[已经过的观察日天数, 已实现的采集价的累加值]；几何平均：[已经过的观察日天数, 已实现的采集价的自然对数累加值]
            self.status = [0, 0]
        else:
            self.status = list(status)
        if engine is not None:
            self.set_pricing_engine(engine)
        else:
            if engine_method is None:
                # 平均结算价，非增强亚式，当几何平均且整个存续期观察，或者算数平均且观察结束日等于终止日期，并且未到平均观察期内时，使用解析解定价引擎
                if (self.substitute == AsianAveSubstitution.Underlying and (not self.enhanced)
                        and self.end_date == self.obs_end and (self.ave_method == AverageMethod.Arithmetic or (
                                self.ave_method == AverageMethod.Geometric and self.start_date == self.obs_start))
                        and global_evaluation_date() <= self.obs_start):
                    default_engine = AnalyticAsianEngine
                else:  # 其他情形：例如替代执行价亚式/增强亚式，使用MC定价引擎
                    default_engine = MCAsianEngine
            elif engine_method == "Analytic":
                default_engine = AnalyticAsianEngine
            elif engine_method in ["MC", "MonteCarlo"]:
                default_engine = MCAsianEngine
            elif engine_method == "TREE":
                default_engine = BiTreeAsianEngine
            else:
                raise NotImplementedError(
                    f"不支持的engine_method{engine_method}, {self.__repr__()}仅支持Analytic/MC/TREE")
            if stoch_process is not None:
                # 未提供引擎，但提供了随机过程
                self.set_pricing_engine(default_engine(stoch_process=stoch_process))
            elif s is not None and r is not None and q is not None and vol is not None:
                # 未提供引擎，但提供了标的价格、无风险利率、分红/融券率、波动率
                self.set_pricing_engine(default_engine(s=s, r=r, q=q, vol=vol))

    def set_pricing_engine(self, engine):
        """设置定价引擎"""
        with suppress(AttributeError):
            self.engine.process.spot.remove_observer(self)
        self.engine = engine
        self.engine.process.spot.add_observer(self)
        logging.info(f"{self}当前定价方法为{engine.engine_type.value}")

    def remove_self(self):
        """删除对象自己，del自己之前，先将自己从被观察者的列表中移除"""
        self.engine.process.spot.remove_observer(self)
        del self

    def update(self, observable, *args, **kwargs):
        """收到标的价格变化的通知时，自动更新亚式期权的已实现平均观察价"""
        if global_auto_update_status():
            calculate_date = global_evaluation_date()
            if calculate_date < self.start_date or calculate_date > self.end_date:
                return  # 如果在起始日前或到期日之后，状态维持不变
            if (observable == self.engine.process.spot
                    and self.obs_start <= calculate_date <= self.obs_end):
                self.status[0] += 1
                if self.enhanced:  # 如果是增强亚式，采集价受到约定价格的限制
                    if self.callput == CallPut.Call:
                        observed_price = max(self.limited_price, observable.data)
                    elif self.callput == CallPut.Put:
                        observed_price = min(self.limited_price, observable.data)
                    else:
                        raise ValueError(f"Error: 不支持的看涨看跌方向{self.callput}")
                else:
                    observed_price = observable.data
                if self.ave_method == AverageMethod.Arithmetic:  # 算术平均
                    self.status[1] += round(self.status[1] + observed_price, 4)
                    logging.info(f"{self} 已实现采集价累加值 更新为: {self.status[1]}, 已观察{self.status[0]}天")
                elif self.ave_method == AverageMethod.Geometric:  # 几何平均
                    self.status[1] = round(self.status[1] + math.log(observed_price), 4)
                    logging.info(f"{self} 已实现采集价的对数累加值 更新为: {self.status[1]}, 已观察{self.status[0]}天")
                else:
                    raise ValueError(f"Error: 未知的平均价计算方法{self.ave_method}")

    def __repr__(self):
        """返回期权的描述"""
        enhance_str = "增强型" if self.enhanced else ""
        return f"{self.ave_method.value}-{self.substitute.value}-{self.callput.name}{enhance_str}亚式期权"

    @time_this
    def price(self, t: datetime.date = None, spot=None):
        """计算期权价格
        Args:
            t: datetime.date，计算期权价格的日期
            spot: float，标的价格
        Returns: 期权现值
        """
        self.validate_parameters(t=t)
        price = self.engine.calc_present_value(prod=self, t=t, spot=spot) * self.parti
        return price
